'use strict';

var internal = require('./internal.cjs');

const slotsMap = new WeakMap();
// TODO: allow type-input, so caller doesn't need to cast so much
const getSlots = slotsMap.get.bind(slotsMap);
const setSlots = slotsMap.set.bind(slotsMap);
function createSlotClass(branding, construct, getters, methods, staticMethods) {
    function Class(...args) {
        if (this instanceof Class) {
            setSlots(this, construct(...args));
        }
        else {
            throw new TypeError(internal.invalidCallingContext);
        }
    }
    Object.defineProperties(Class.prototype, {
        ...internal.createGetterDescriptors(internal.mapProps(bindMethod, getters)), // !!!
        ...internal.createPropDescriptors(internal.mapProps(bindMethod, methods)),
        ...internal.createStringTagDescriptors('Temporal.' + branding),
    });
    Object.defineProperties(Class, {
        ...internal.createPropDescriptors(staticMethods),
        ...internal.createNameDescriptors(branding),
    });
    function bindMethod(method, methodName) {
        return Object.defineProperties(function (...args) {
            return method.call(this, getSpecificSlots(this), ...args);
        }, internal.createNameDescriptors(methodName));
    }
    function getSpecificSlots(obj) {
        const slots = getSlots(obj);
        if (!slots || slots.branding !== branding) {
            throw new TypeError(internal.invalidCallingContext);
        }
        return slots;
    }
    function createViaSlots(slots) {
        const instance = Object.create(Class.prototype);
        setSlots(instance, slots);
        return instance;
    }
    return [Class, createViaSlots, getSpecificSlots];
}
// Utils
// -----------------------------------------------------------------------------
// TODO: return type
function createProtocolValidator(propNames) {
    propNames = propNames.concat('id').sort();
    return (obj) => {
        if (!internal.hasAllPropsByName(obj, propNames)) {
            throw new TypeError(internal.invalidProtocol);
        }
        return obj;
    };
}
function rejectInvalidBag(bag) {
    if (getSlots(bag) ||
        bag.calendar !== undefined ||
        bag.timeZone !== undefined) {
        throw new TypeError(internal.invalidBag);
    }
    return bag;
}

const DateTimeFormat = createDateTimeFormatClass();
const internalsMap = new WeakMap();
function createDateTimeFormatClass() {
    const members = internal.RawDateTimeFormat.prototype;
    const memberDescriptors = Object.getOwnPropertyDescriptors(members);
    const classDescriptors = Object.getOwnPropertyDescriptors(internal.RawDateTimeFormat);
    const DateTimeFormat = function (locales, options = {}) {
        // Constructor can be called without `new`
        // https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/DateTimeFormat/DateTimeFormat#return_value
        if (!(this instanceof DateTimeFormat)) {
            return new DateTimeFormat(locales, options);
        }
        internalsMap.set(this, createDateTimeFormatInternals(locales, options));
    };
    for (const memberName in memberDescriptors) {
        const memberDescriptor = memberDescriptors[memberName];
        const formatLikeMethod = memberName.startsWith('format') && createFormatMethod(memberName);
        if (typeof memberDescriptor.value === 'function') {
            memberDescriptor.value =
                memberName === 'constructor'
                    ? DateTimeFormat
                    : formatLikeMethod || createProxiedMethod(memberName);
        }
        else if (formatLikeMethod) {
            // .format() is always bound to the instance. It's a getter
            // https://tc39.es/ecma402/#sec-intl.datetimeformat.prototype.format
            memberDescriptor.get = function () {
                return formatLikeMethod.bind(this);
            };
        }
    }
    classDescriptors.prototype.value = Object.create(members, memberDescriptors);
    Object.defineProperties(DateTimeFormat, classDescriptors);
    return DateTimeFormat;
}
function createFormatMethod(methodName) {
    return function (...formattables) {
        const prepFormat = internalsMap.get(this);
        const [format, ...rawFormattables] = prepFormat(...formattables);
        return format[methodName](...rawFormattables);
    };
}
function createProxiedMethod(methodName) {
    return function (...args) {
        const prepFormat = internalsMap.get(this);
        return prepFormat.rawFormat[methodName](...args);
    };
}
// Config
// -----------------------------------------------------------------------------
const classFormatConfigs = {
    PlainYearMonth: internal.plainYearMonthConfig,
    PlainMonthDay: internal.plainMonthDayConfig,
    PlainDate: internal.plainDateConfig,
    PlainDateTime: internal.plainDateTimeConfig,
    PlainTime: internal.plainTimeConfig,
    Instant: internal.instantConfig,
    // ZonedDateTime not allowed to be formatted by Intl.DateTimeFormat
};
function createDateTimeFormatInternals(locales, options = {}) {
    const rawFormat = new internal.RawDateTimeFormat(locales, options);
    const resolveOptions = rawFormat.resolvedOptions();
    const resolvedLocale = resolveOptions.locale;
    const copiedOptions = internal.pluckProps(Object.keys(options), resolveOptions);
    const queryFormatPrepperForBranding = internal.createLazyGenerator(createFormatPrepperForBranding);
    const prepFormat = (...formattables) => {
        let branding;
        const slotsList = formattables.map((formattable, i) => {
            const slots = getSlots(formattable);
            const slotsBranding = (slots || {}).branding;
            if (i && branding && branding !== slotsBranding) {
                throw new TypeError(internal.mismatchingFormatTypes);
            }
            branding = slotsBranding;
            return slots;
        });
        if (branding) {
            return queryFormatPrepperForBranding(branding)(resolvedLocale, copiedOptions, ...slotsList);
        }
        return [rawFormat, ...formattables];
    };
    prepFormat.rawFormat = rawFormat;
    return prepFormat;
}
function createFormatPrepperForBranding(branding) {
    const config = classFormatConfigs[branding];
    if (!config) {
        throw new TypeError(internal.invalidFormatType(branding));
    }
    return internal.createFormatPrepper(config, 
    // a generator that conveniently caches by the first arg: forcedTimeZoneId
    internal.createLazyGenerator(internal.createFormatForPrep));
}
// Format Prepping for each class' toLocaleString
// (best place for this?)
// -----------------------------------------------------------------------------
const prepPlainYearMonthFormat = internal.createFormatPrepper(internal.plainYearMonthConfig);
const prepPlainMonthDayFormat = internal.createFormatPrepper(internal.plainMonthDayConfig);
const prepPlainDateFormat = internal.createFormatPrepper(internal.plainDateConfig);
const prepPlainDateTimeFormat = internal.createFormatPrepper(internal.plainDateTimeConfig);
const prepPlainTimeFormat = internal.createFormatPrepper(internal.plainTimeConfig);
const prepInstantFormat = internal.createFormatPrepper(internal.instantConfig);
const prepZonedDateTimeFormat = internal.createFormatPrepper(internal.zonedDateTimeConfig);

const yearMonthOnlyRefiners = {
    era: internal.requireStringOrUndefined,
    eraYear: internal.requireIntegerOrUndefined,
    year: internal.requireInteger,
    month: internal.requirePositiveInteger,
    daysInMonth: internal.requirePositiveInteger,
    daysInYear: internal.requirePositiveInteger,
    inLeapYear: internal.requireBoolean,
    monthsInYear: internal.requirePositiveInteger,
};
const monthOnlyRefiners = {
    monthCode: internal.requireString,
};
const dayOnlyRefiners = {
    day: internal.requirePositiveInteger,
};
const dateOnlyRefiners = {
    dayOfWeek: internal.requirePositiveInteger,
    dayOfYear: internal.requirePositiveInteger,
    weekOfYear: internal.requirePositiveInteger,
    yearOfWeek: internal.requireInteger,
    daysInWeek: internal.requirePositiveInteger,
};
const dateRefiners = {
    ...yearMonthOnlyRefiners,
    ...monthOnlyRefiners,
    ...dayOnlyRefiners,
    ...dateOnlyRefiners,
};

// For Calendar
// -----------------------------------------------------------------------------
// Always assumes underlying Native calendar `ops`
function createCalendarFieldMethods(methodNameMap, alsoAccept) {
    const methods = {};
    for (const methodName in methodNameMap) {
        methods[methodName] = function ({ native }, dateArg) {
            const argSlots = (getSlots(dateArg) || {});
            const { branding } = argSlots;
            const refinedSlots = branding === internal.PlainDateBranding || alsoAccept.includes(branding)
                ? argSlots
                : toPlainDateSlots(dateArg);
            return native[methodName](refinedSlots);
        };
    }
    return methods;
}
const calendarFieldMethods = {
    ...createCalendarFieldMethods(yearMonthOnlyRefiners, [
        internal.PlainYearMonthBranding,
    ]),
    ...createCalendarFieldMethods(dateOnlyRefiners, []),
    ...createCalendarFieldMethods(monthOnlyRefiners, [
        internal.PlainYearMonthBranding,
        internal.PlainMonthDayBranding,
    ]),
    ...createCalendarFieldMethods(dayOnlyRefiners, [internal.PlainMonthDayBranding]),
};
// For PlainDate/etc
// -----------------------------------------------------------------------------
// Assumes general calendar (native/adapter)
function createCalendarGetters(methodNameMap) {
    const methods = {};
    for (const methodName in methodNameMap) {
        methods[methodName] = function (slots) {
            const { calendar } = slots;
            const simpleOps = createSimpleOps(calendar);
            return simpleOps[methodName](slots);
        };
    }
    return methods;
}
const dateGetters = createCalendarGetters(dateRefiners);
const yearMonthGetters = createCalendarGetters({
    ...yearMonthOnlyRefiners,
    ...monthOnlyRefiners,
});
const monthDayGetters = createCalendarGetters({
    ...monthOnlyRefiners,
    ...dayOnlyRefiners,
});
const calendarIdGetters = {
    calendarId(slots) {
        return internal.getId(slots.calendar);
    },
};
const adapterSimpleOps = internal.mapProps((refiner, methodName) => {
    return function (isoFields) {
        const { calendarProtocol } = this;
        return refiner(calendarProtocol[methodName](createPlainDate(internal.createPlainDateSlots(isoFields, calendarProtocol))));
    };
}, dateRefiners);
function createAdapterSimpleOps(calendarProtocol) {
    return Object.assign(Object.create(adapterSimpleOps), {
        calendarProtocol,
    });
}
function createSimpleOps(calendarSlot) {
    if (typeof calendarSlot === 'string') {
        return internal.createNativeStandardOps(calendarSlot); // has everything
    }
    return createAdapterSimpleOps(calendarSlot);
}
// Duration
// -----------------------------------------------------------------------------
const durationGetters = internal.mapPropNames((propName) => {
    return function (slots) {
        return slots[propName];
    };
}, internal.durationFieldNamesAsc);
// Time
// -----------------------------------------------------------------------------
const timeGetters = internal.mapPropNames((_name, i) => {
    return function (slots) {
        return slots[internal.isoTimeFieldNamesAsc[i]];
    };
}, internal.timeFieldNamesAsc);
// Epoch
// -----------------------------------------------------------------------------
const epochGetters = {
    epochSeconds: internal.getEpochSeconds,
    epochMilliseconds: internal.getEpochMilliseconds,
    epochMicroseconds: internal.getEpochMicroseconds,
    epochNanoseconds: internal.getEpochNanoseconds,
};
// Misc
// -----------------------------------------------------------------------------
function neverValueOf() {
    throw new TypeError(internal.forbiddenValueOf);
}
function createCalendarFromSlots({ calendar, }) {
    return typeof calendar === 'string' ? new Calendar(calendar) : calendar;
}
// for getISOFields
const removeBranding = internal.bindArgs(internal.excludePropsByName, new Set([/*@__KEY__*/ 'branding']));

const [PlainMonthDay, createPlainMonthDay, getPlainMonthDaySlots] = createSlotClass(internal.PlainMonthDayBranding, internal.bindArgs(internal.constructPlainMonthDaySlots, refineCalendarSlot), {
    ...calendarIdGetters,
    ...monthDayGetters,
}, {
    with(slots, mod, options) {
        return createPlainMonthDay(internal.plainMonthDayWithFields(createMonthDayModOps, slots, this, rejectInvalidBag(mod), options));
    },
    equals(slots, otherArg) {
        return internal.plainMonthDaysEqual(slots, toPlainMonthDaySlots(otherArg));
    },
    toString: internal.formatPlainMonthDayIso,
    toJSON(slots) {
        return internal.formatPlainMonthDayIso(slots);
    },
    toLocaleString(slots, locales, options) {
        const [format, epochMilli] = prepPlainMonthDayFormat(locales, options, slots);
        return format.format(epochMilli);
    },
    toPlainDate(slots, bag) {
        return createPlainDate(internal.plainMonthDayToPlainDate(createDateModOps, slots, this, bag));
    },
    getISOFields: removeBranding,
    getCalendar: createCalendarFromSlots,
    valueOf: neverValueOf,
}, {
    from(arg, options) {
        return createPlainMonthDay(toPlainMonthDaySlots(arg, options));
    },
});
// Utils
// -----------------------------------------------------------------------------
function toPlainMonthDaySlots(arg, options) {
    options = internal.copyOptions(options);
    if (internal.isObjectLike(arg)) {
        const slots = getSlots(arg);
        if (slots && slots.branding === internal.PlainMonthDayBranding) {
            internal.refineOverflowOptions(options); // parse unused options
            return slots;
        }
        const calendarMaybe = extractCalendarSlotFromBag(arg);
        const calendar = calendarMaybe || internal.isoCalendarId;
        return internal.refinePlainMonthDayBag(createMonthDayRefineOps(calendar), !calendarMaybe, arg, options);
    }
    const res = internal.parsePlainMonthDay(internal.createNativeStandardOps, arg);
    internal.refineOverflowOptions(options); // parse unused options
    return res;
}

// Individual Adapters
// -----------------------------------------------------------------------------
function getOffsetNanosecondsForAdapter(timeZoneProtocol, getOffsetNanosecondsFor, epochNano) {
    return validateTimeZoneOffsetRes(getOffsetNanosecondsFor.call(timeZoneProtocol, createInstant(internal.createInstantSlots(epochNano))));
}
function getPossibleInstantsForAdapter(timeZoneProtocol, getPossibleInstantsFor, isoFields) {
    return [
        ...getPossibleInstantsFor.call(timeZoneProtocol, createPlainDateTime(internal.createPlainDateTimeSlots(isoFields, internal.isoCalendarId))),
    ].map((instant) => {
        return getInstantSlots(instant).epochNanoseconds;
    });
}
function validateTimeZoneOffsetRes(offsetNano) {
    return internal.validateTimeZoneOffset(internal.requireInteger(offsetNano));
}
// Adapter Sets
// -----------------------------------------------------------------------------
const timeZoneAdapters = {
    getOffsetNanosecondsFor: getOffsetNanosecondsForAdapter,
    getPossibleInstantsFor: getPossibleInstantsForAdapter,
};
// TODO: rename to be about 'offset'
const simpleTimeZoneAdapters = {
    getOffsetNanosecondsFor: getOffsetNanosecondsForAdapter,
};
function createAdapterOps(timeZoneProtocol, adapterFuncs = timeZoneAdapters) {
    const keys = Object.keys(adapterFuncs).sort();
    const boundFuncs = {};
    for (const key of keys) {
        boundFuncs[key] = internal.bindArgs(adapterFuncs[key], timeZoneProtocol, internal.requireFunction(timeZoneProtocol[key]));
    }
    return boundFuncs;
}

function createTimeZoneOps(timeZoneSlot, adapterFuncs) {
    if (typeof timeZoneSlot === 'string') {
        return internal.queryNativeTimeZone(timeZoneSlot);
    }
    return createAdapterOps(timeZoneSlot, adapterFuncs);
}
function createTimeZoneOffsetOps(timeZoneSlot) {
    return createTimeZoneOps(timeZoneSlot, simpleTimeZoneAdapters);
}

const [PlainYearMonth, createPlainYearMonth, getPlainYearMonthSlots] = createSlotClass(internal.PlainYearMonthBranding, internal.bindArgs(internal.constructPlainYearMonthSlots, refineCalendarSlot), {
    ...calendarIdGetters,
    ...yearMonthGetters,
}, {
    with(slots, mod, options) {
        return createPlainYearMonth(internal.plainYearMonthWithFields(createYearMonthModOps, slots, this, rejectInvalidBag(mod), options));
    },
    add(slots, durationArg, options) {
        return createPlainYearMonth(internal.movePlainYearMonth(createYearMonthMoveOps, false, slots, toDurationSlots(durationArg), options));
    },
    subtract(slots, durationArg, options) {
        return createPlainYearMonth(internal.movePlainYearMonth(createYearMonthMoveOps, true, slots, toDurationSlots(durationArg), options));
    },
    until(slots, otherArg, options) {
        return createDuration(internal.diffPlainYearMonth(createYearMonthDiffOps, false, slots, toPlainYearMonthSlots(otherArg), options));
    },
    since(slots, otherArg, options) {
        return createDuration(internal.diffPlainYearMonth(createYearMonthDiffOps, true, slots, toPlainYearMonthSlots(otherArg), options));
    },
    equals(slots, otherArg) {
        return internal.plainYearMonthsEqual(slots, toPlainYearMonthSlots(otherArg));
    },
    toString: internal.formatPlainYearMonthIso,
    toJSON(slots) {
        return internal.formatPlainYearMonthIso(slots);
    },
    toLocaleString(slots, locales, options) {
        const [format, epochMilli] = prepPlainYearMonthFormat(locales, options, slots);
        return format.format(epochMilli);
    },
    toPlainDate(slots, bag) {
        return createPlainDate(internal.plainYearMonthToPlainDate(createDateModOps, slots, this, bag));
    },
    getISOFields: removeBranding,
    getCalendar: createCalendarFromSlots,
    valueOf: neverValueOf,
}, {
    from(arg, options) {
        return createPlainYearMonth(toPlainYearMonthSlots(arg, options));
    },
    compare(arg0, arg1) {
        return internal.compareIsoDateFields(toPlainYearMonthSlots(arg0), toPlainYearMonthSlots(arg1));
    },
});
// Utils
// -----------------------------------------------------------------------------
function toPlainYearMonthSlots(arg, options) {
    options = internal.copyOptions(options);
    if (internal.isObjectLike(arg)) {
        const slots = getSlots(arg);
        if (slots && slots.branding === internal.PlainYearMonthBranding) {
            internal.refineOverflowOptions(options); // parse unused options
            return slots;
        }
        return internal.refinePlainYearMonthBag(createYearMonthRefineOps(getCalendarSlotFromBag(arg)), // !!!
        arg, // !!!
        options);
    }
    const res = internal.parsePlainYearMonth(internal.createNativeStandardOps, arg);
    internal.refineOverflowOptions(options); // parse unused options
    return res;
}

const [ZonedDateTime, createZonedDateTime] = createSlotClass(internal.ZonedDateTimeBranding, internal.bindArgs(internal.constructZonedDateTimeSlots, refineCalendarSlot, refineTimeZoneSlot), {
    ...epochGetters,
    ...calendarIdGetters,
    ...adaptDateMethods(dateGetters),
    ...adaptDateMethods(timeGetters),
    hoursInDay(slots) {
        return internal.computeHoursInDay(createTimeZoneOps, slots);
    },
    offsetNanoseconds(slots) {
        return slotsToIso(slots).offsetNanoseconds;
    },
    offset(slots) {
        return internal.formatOffsetNano(slotsToIso(slots).offsetNanoseconds);
    },
    timeZoneId(slots) {
        return internal.getId(slots.timeZone);
    },
}, {
    with(slots, mod, options) {
        return createZonedDateTime(internal.zonedDateTimeWithFields(createDateModOps, createTimeZoneOps, slots, this, rejectInvalidBag(mod), options));
    },
    withPlainTime(slots, plainTimeArg) {
        return createZonedDateTime(internal.zonedDateTimeWithPlainTime(createTimeZoneOps, slots, optionalToPlainTimeFields(plainTimeArg)));
    },
    withPlainDate(slots, plainDateArg) {
        return createZonedDateTime(internal.zonedDateTimeWithPlainDate(createTimeZoneOps, slots, toPlainDateSlots(plainDateArg)));
    },
    withTimeZone(slots, timeZoneArg) {
        return createZonedDateTime(internal.slotsWithTimeZone(slots, refineTimeZoneSlot(timeZoneArg)));
    },
    withCalendar(slots, calendarArg) {
        return createZonedDateTime(internal.slotsWithCalendar(slots, refineCalendarSlot(calendarArg)));
    },
    add(slots, durationArg, options) {
        return createZonedDateTime(internal.moveZonedDateTime(createMoveOps, createTimeZoneOps, false, slots, toDurationSlots(durationArg), options));
    },
    subtract(slots, durationArg, options) {
        return createZonedDateTime(internal.moveZonedDateTime(createMoveOps, createTimeZoneOps, true, slots, toDurationSlots(durationArg), options));
    },
    until(slots, otherArg, options) {
        return createDuration(internal.createDurationSlots(internal.diffZonedDateTimes(createDiffOps, createTimeZoneOps, false, slots, toZonedDateTimeSlots(otherArg), options)));
    },
    since(slots, otherArg, options) {
        return createDuration(internal.createDurationSlots(internal.diffZonedDateTimes(createDiffOps, createTimeZoneOps, true, slots, toZonedDateTimeSlots(otherArg), options)));
    },
    round(slots, options) {
        return createZonedDateTime(internal.roundZonedDateTime(createTimeZoneOps, slots, options));
    },
    startOfDay(slots) {
        return createZonedDateTime(internal.computeStartOfDay(createTimeZoneOps, slots));
    },
    equals(slots, otherArg) {
        return internal.zonedDateTimesEqual(slots, toZonedDateTimeSlots(otherArg));
    },
    toString(slots, options) {
        return internal.formatZonedDateTimeIso(createTimeZoneOffsetOps, slots, options);
    },
    toJSON(slots) {
        return internal.formatZonedDateTimeIso(createTimeZoneOffsetOps, slots);
    },
    toLocaleString(slots, locales, options = {}) {
        const [format, epochMilli] = prepZonedDateTimeFormat(locales, options, slots);
        return format.format(epochMilli);
    },
    toInstant(slots) {
        return createInstant(internal.zonedDateTimeToInstant(slots));
    },
    toPlainDate(slots) {
        return createPlainDate(internal.zonedDateTimeToPlainDate(createTimeZoneOffsetOps, slots));
    },
    toPlainTime(slots) {
        return createPlainTime(internal.zonedDateTimeToPlainTime(createTimeZoneOffsetOps, slots));
    },
    toPlainDateTime(slots) {
        return createPlainDateTime(internal.zonedDateTimeToPlainDateTime(createTimeZoneOffsetOps, slots));
    },
    toPlainYearMonth(slots) {
        return createPlainYearMonth(internal.zonedDateTimeToPlainYearMonth(createYearMonthRefineOps, slots, this));
    },
    toPlainMonthDay(slots) {
        return createPlainMonthDay(internal.zonedDateTimeToPlainMonthDay(createMonthDayRefineOps, slots, this));
    },
    getISOFields(slots) {
        return internal.buildZonedIsoFields(createTimeZoneOffsetOps, slots);
    },
    getCalendar: createCalendarFromSlots,
    getTimeZone({ timeZone, }) {
        return typeof timeZone === 'string' ? new TimeZone(timeZone) : timeZone;
    },
    valueOf: neverValueOf,
}, {
    from(arg, options) {
        return createZonedDateTime(toZonedDateTimeSlots(arg, options));
    },
    compare(arg0, arg1) {
        return internal.compareZonedDateTimes(toZonedDateTimeSlots(arg0), toZonedDateTimeSlots(arg1));
    },
});
// Utils
// -----------------------------------------------------------------------------
function toZonedDateTimeSlots(arg, options) {
    options = internal.copyOptions(options);
    if (internal.isObjectLike(arg)) {
        const slots = getSlots(arg);
        if (slots && slots.branding === internal.ZonedDateTimeBranding) {
            internal.refineZonedFieldOptions(options); // parse unused options
            return slots;
        }
        const calendarSlot = getCalendarSlotFromBag(arg);
        return internal.refineZonedDateTimeBag(refineTimeZoneSlot, createTimeZoneOps, createDateRefineOps(calendarSlot), calendarSlot, arg, // !!!
        options);
    }
    return internal.parseZonedDateTime(arg, options);
}
function slotsToIso(slots) {
    return internal.zonedEpochSlotsToIso(slots, createTimeZoneOffsetOps);
}
function adaptDateMethods(methods) {
    return internal.mapProps((method) => {
        return (slots) => {
            return method(slotsToIso(slots));
        };
    }, methods);
}

const [Instant, createInstant, getInstantSlots] = createSlotClass(internal.InstantBranding, internal.constructInstantSlots, epochGetters, {
    add(slots, durationArg) {
        return createInstant(internal.moveInstant(false, slots, toDurationSlots(durationArg)));
    },
    subtract(slots, durationArg) {
        return createInstant(internal.moveInstant(true, slots, toDurationSlots(durationArg)));
    },
    until(slots, otherArg, options) {
        return createDuration(internal.diffInstants(false, slots, toInstantSlots(otherArg), options));
    },
    since(slots, otherArg, options) {
        return createDuration(internal.diffInstants(true, slots, toInstantSlots(otherArg), options));
    },
    round(slots, options) {
        return createInstant(internal.roundInstant(slots, options));
    },
    equals(slots, otherArg) {
        return internal.instantsEqual(slots, toInstantSlots(otherArg));
    },
    toString(slots, options) {
        return internal.formatInstantIso(refineTimeZoneSlot, createTimeZoneOffsetOps, slots, options);
    },
    toJSON(slots) {
        return internal.formatInstantIso(refineTimeZoneSlot, createTimeZoneOffsetOps, slots);
    },
    toLocaleString(slots, locales, options) {
        const [format, epochMilli] = prepInstantFormat(locales, options, slots);
        return format.format(epochMilli);
    },
    toZonedDateTimeISO(slots, timeZoneArg) {
        return createZonedDateTime(internal.instantToZonedDateTime(slots, refineTimeZoneSlot(timeZoneArg)));
    },
    toZonedDateTime(slots, options) {
        const refinedObj = internal.requireObjectLike(options);
        return createZonedDateTime(internal.instantToZonedDateTime(slots, refineTimeZoneSlot(refinedObj.timeZone), refineCalendarSlot(refinedObj.calendar)));
    },
    valueOf: neverValueOf,
}, {
    from(arg) {
        return createInstant(toInstantSlots(arg));
    },
    fromEpochSeconds(epochSec) {
        return createInstant(internal.epochSecToInstant(epochSec));
    },
    fromEpochMilliseconds(epochMilli) {
        return createInstant(internal.epochMilliToInstant(epochMilli));
    },
    fromEpochMicroseconds(epochMicro) {
        return createInstant(internal.epochMicroToInstant(epochMicro));
    },
    fromEpochNanoseconds(epochNano) {
        return createInstant(internal.epochNanoToInstant(epochNano));
    },
    compare(a, b) {
        return internal.compareInstants(toInstantSlots(a), toInstantSlots(b));
    },
});
// Utils
// -----------------------------------------------------------------------------
function toInstantSlots(arg) {
    if (internal.isObjectLike(arg)) {
        const slots = getSlots(arg);
        if (slots) {
            switch (slots.branding) {
                case internal.InstantBranding:
                    return slots;
                case internal.ZonedDateTimeBranding:
                    return internal.createInstantSlots(slots
                        .epochNanoseconds);
            }
        }
    }
    return internal.parseInstant(arg);
}
// Legacy Date
// -----------------------------------------------------------------------------
function toTemporalInstant() {
    return createInstant(internal.createInstantSlots(internal.numberToDayTimeNano(this.valueOf(), internal.nanoInMilli)));
}

const [TimeZone, createTimeZone] = createSlotClass('TimeZone', (id) => {
    const slotId = internal.resolveTimeZoneId(id);
    const timeZoneNative = internal.queryNativeTimeZone(slotId);
    return {
        branding: 'TimeZone',
        id: slotId,
        native: timeZoneNative,
    };
}, {
    id(slots) {
        return slots.id;
    },
}, {
    toString(slots) {
        return slots.id;
    },
    toJSON(slots) {
        return slots.id;
    },
    getPossibleInstantsFor({ native }, plainDateTimeArg) {
        return native
            .getPossibleInstantsFor(toPlainDateTimeSlots(plainDateTimeArg))
            .map((epochNano) => {
            return createInstant(internal.createInstantSlots(epochNano));
        });
    },
    getOffsetNanosecondsFor({ native }, instantArg) {
        return native.getOffsetNanosecondsFor(toInstantSlots(instantArg).epochNanoseconds);
    },
    getOffsetStringFor(_slots, instantArg) {
        const epochNano = toInstantSlots(instantArg).epochNanoseconds;
        const calendarOps = createAdapterOps(this, simpleTimeZoneAdapters); // for accessing own methods
        const offsetNano = calendarOps.getOffsetNanosecondsFor(epochNano);
        return internal.formatOffsetNano(offsetNano);
    },
    getPlainDateTimeFor(_slots, instantArg, calendarArg = internal.isoCalendarId) {
        const epochNano = toInstantSlots(instantArg).epochNanoseconds;
        const calendarOps = createAdapterOps(this, simpleTimeZoneAdapters); // for accessing own methods
        const offsetNano = calendarOps.getOffsetNanosecondsFor(epochNano);
        return createPlainDateTime(internal.createPlainDateTimeSlots(internal.epochNanoToIso(epochNano, offsetNano), refineCalendarSlot(calendarArg)));
    },
    getInstantFor(_slots, plainDateTimeArg, options) {
        const isoFields = toPlainDateTimeSlots(plainDateTimeArg);
        const epochDisambig = internal.refineEpochDisambigOptions(options);
        const calendarOps = createAdapterOps(this); // for accessing own methods
        return createInstant(internal.createInstantSlots(internal.getSingleInstantFor(calendarOps, isoFields, epochDisambig)));
    },
    getNextTransition({ native }, instantArg) {
        return getImplTransition(1, native, instantArg);
    },
    getPreviousTransition({ native }, instantArg) {
        return getImplTransition(-1, native, instantArg);
    },
    equals(_slots, otherArg) {
        return !!internal.isTimeZoneSlotsEqual(this, refineTimeZoneSlot(otherArg));
    },
}, {
    from(arg) {
        const timeZoneSlot = refineTimeZoneSlot(arg);
        return typeof timeZoneSlot === 'string'
            ? new TimeZone(timeZoneSlot)
            : timeZoneSlot;
    },
});
// Utils
// -----------------------------------------------------------------------------
function getImplTransition(direction, impl, instantArg) {
    const epochNano = impl.getTransition(toInstantSlots(instantArg).epochNanoseconds, direction);
    return epochNano ? createInstant(internal.createInstantSlots(epochNano)) : null;
}
function refineTimeZoneSlot(arg) {
    if (internal.isObjectLike(arg)) {
        return ((getSlots(arg) || {}).timeZone ||
            validateTimeZoneProtocol(arg));
    }
    return refineTimeZoneSlotString(arg);
}
function refineTimeZoneSlotString(arg) {
    return internal.resolveTimeZoneId(internal.parseTimeZoneId(internal.requireString(arg)));
}
const validateTimeZoneProtocol = createProtocolValidator(Object.keys(timeZoneAdapters));

const [PlainTime, createPlainTime] = createSlotClass(internal.PlainTimeBranding, internal.constructPlainTimeSlots, timeGetters, {
    with(_slots, mod, options) {
        return createPlainTime(internal.plainTimeWithFields(this, rejectInvalidBag(mod), options));
    },
    add(slots, durationArg) {
        return createPlainTime(internal.movePlainTime(false, slots, toDurationSlots(durationArg)));
    },
    subtract(slots, durationArg) {
        return createPlainTime(internal.movePlainTime(true, slots, toDurationSlots(durationArg)));
    },
    until(slots, otherArg, options) {
        return createDuration(internal.diffPlainTimes(false, slots, toPlainTimeSlots(otherArg), options));
    },
    since(slots, otherArg, options) {
        return createDuration(internal.diffPlainTimes(true, slots, toPlainTimeSlots(otherArg), options));
    },
    round(slots, options) {
        return createPlainTime(internal.roundPlainTime(slots, options));
    },
    equals(slots, other) {
        return internal.plainTimesEqual(slots, toPlainTimeSlots(other));
    },
    toString: internal.formatPlainTimeIso,
    toJSON(slots) {
        return internal.formatPlainTimeIso(slots);
    },
    toLocaleString(slots, locales, options) {
        const [format, epochMilli] = prepPlainTimeFormat(locales, options, slots);
        return format.format(epochMilli);
    },
    toZonedDateTime(slots, options) {
        return createZonedDateTime(internal.plainTimeToZonedDateTime(refineTimeZoneSlot, toPlainDateSlots, createTimeZoneOps, slots, options));
    },
    toPlainDateTime(slots, plainDateArg) {
        return createPlainDateTime(internal.plainTimeToPlainDateTime(slots, toPlainDateSlots(plainDateArg)));
    },
    getISOFields: removeBranding,
    valueOf: neverValueOf,
}, {
    from(arg, options) {
        return createPlainTime(toPlainTimeSlots(arg, options));
    },
    compare(arg0, arg1) {
        return internal.compareIsoTimeFields(toPlainTimeSlots(arg0), toPlainTimeSlots(arg1));
    },
});
// Utils
// -----------------------------------------------------------------------------
function toPlainTimeSlots(arg, options) {
    if (internal.isObjectLike(arg)) {
        const slots = (getSlots(arg) || {});
        switch (slots.branding) {
            case internal.PlainTimeBranding:
                internal.refineOverflowOptions(options); // parse unused options
                return slots;
            case internal.PlainDateTimeBranding:
                internal.refineOverflowOptions(options); // parse unused options
                return internal.createPlainTimeSlots(slots);
            case internal.ZonedDateTimeBranding:
                internal.refineOverflowOptions(options); // parse unused options
                return internal.zonedDateTimeToPlainTime(createTimeZoneOffsetOps, slots);
        }
        return internal.refinePlainTimeBag(arg, options);
    }
    internal.refineOverflowOptions(options); // parse unused options
    return internal.parsePlainTime(arg);
}
function optionalToPlainTimeFields(timeArg) {
    return timeArg === undefined ? undefined : toPlainTimeSlots(timeArg);
}

const [PlainDateTime, createPlainDateTime] = createSlotClass(internal.PlainDateTimeBranding, internal.bindArgs(internal.constructPlainDateTimeSlots, refineCalendarSlot), {
    ...calendarIdGetters,
    ...dateGetters,
    ...timeGetters,
}, {
    with(slots, mod, options) {
        return createPlainDateTime(internal.plainDateTimeWithFields(createDateModOps, slots, this, rejectInvalidBag(mod), options));
    },
    withPlainTime(slots, plainTimeArg) {
        return createPlainDateTime(internal.plainDateTimeWithPlainTime(slots, optionalToPlainTimeFields(plainTimeArg)));
    },
    withPlainDate(slots, plainDateArg) {
        return createPlainDateTime(internal.plainDateTimeWithPlainDate(slots, toPlainDateSlots(plainDateArg)));
    },
    withCalendar(slots, calendarArg) {
        return createPlainDateTime(internal.slotsWithCalendar(slots, refineCalendarSlot(calendarArg)));
    },
    add(slots, durationArg, options) {
        return createPlainDateTime(internal.movePlainDateTime(createMoveOps, false, slots, toDurationSlots(durationArg), options));
    },
    subtract(slots, durationArg, options) {
        return createPlainDateTime(internal.movePlainDateTime(createMoveOps, true, slots, toDurationSlots(durationArg), options));
    },
    until(slots, otherArg, options) {
        return createDuration(internal.diffPlainDateTimes(createDiffOps, false, slots, toPlainDateTimeSlots(otherArg), options));
    },
    since(slots, otherArg, options) {
        return createDuration(internal.diffPlainDateTimes(createDiffOps, true, slots, toPlainDateTimeSlots(otherArg), options));
    },
    round(slots, options) {
        return createPlainDateTime(internal.roundPlainDateTime(slots, options));
    },
    equals(slots, otherArg) {
        return internal.plainDateTimesEqual(slots, toPlainDateTimeSlots(otherArg));
    },
    toString(slots, options) {
        return internal.formatPlainDateTimeIso(slots, options);
    },
    toJSON(slots) {
        return internal.formatPlainDateTimeIso(slots);
    },
    toLocaleString(slots, locales, options) {
        const [format, epochMilli] = prepPlainDateTimeFormat(locales, options, slots);
        return format.format(epochMilli);
    },
    toZonedDateTime(slots, timeZoneArg, options) {
        return createZonedDateTime(internal.plainDateTimeToZonedDateTime(createTimeZoneOps, slots, refineTimeZoneSlot(timeZoneArg), options));
    },
    toPlainDate(slots) {
        return createPlainDate(internal.createPlainDateSlots(slots));
    },
    toPlainYearMonth(slots) {
        return createPlainYearMonth(internal.plainDateTimeToPlainYearMonth(createYearMonthRefineOps, slots, this));
    },
    toPlainMonthDay(slots) {
        return createPlainMonthDay(internal.plainDateTimeToPlainMonthDay(createMonthDayRefineOps, slots, this));
    },
    toPlainTime(slots) {
        return createPlainTime(internal.createPlainTimeSlots(slots));
    },
    getISOFields: removeBranding,
    getCalendar: createCalendarFromSlots,
    valueOf: neverValueOf,
}, {
    from(arg, options) {
        return createPlainDateTime(toPlainDateTimeSlots(arg, options));
    },
    compare(arg0, arg1) {
        return internal.compareIsoDateTimeFields(toPlainDateTimeSlots(arg0), toPlainDateTimeSlots(arg1));
    },
});
// Utils
// -----------------------------------------------------------------------------
function toPlainDateTimeSlots(arg, options) {
    options = internal.copyOptions(options);
    if (internal.isObjectLike(arg)) {
        const slots = (getSlots(arg) || {});
        switch (slots.branding) {
            case internal.PlainDateTimeBranding:
                internal.refineOverflowOptions(options); // parse unused options
                return slots;
            case internal.PlainDateBranding:
                internal.refineOverflowOptions(options); // parse unused options
                return internal.createPlainDateTimeSlots({
                    ...slots,
                    ...internal.isoTimeFieldDefaults,
                });
            case internal.ZonedDateTimeBranding:
                internal.refineOverflowOptions(options); // parse unused options
                return internal.zonedDateTimeToPlainDateTime(createTimeZoneOffsetOps, slots);
        }
        return internal.refinePlainDateTimeBag(createDateRefineOps(getCalendarSlotFromBag(arg)), arg, options);
    }
    const res = internal.parsePlainDateTime(arg);
    internal.refineOverflowOptions(options); // parse unused options
    return res;
}

// TODO: give `this` a type
const [PlainDate, createPlainDate, getPlainDateSlots] = createSlotClass(internal.PlainDateBranding, internal.bindArgs(internal.constructPlainDateSlots, refineCalendarSlot), {
    ...calendarIdGetters,
    ...dateGetters,
}, {
    with(slots, mod, options) {
        return createPlainDate(internal.plainDateWithFields(createDateModOps, slots, this, rejectInvalidBag(mod), options));
    },
    withCalendar(slots, calendarArg) {
        return createPlainDate(internal.slotsWithCalendar(slots, refineCalendarSlot(calendarArg)));
    },
    add(slots, durationArg, options) {
        return createPlainDate(internal.movePlainDate(createMoveOps, false, slots, toDurationSlots(durationArg), options));
    },
    subtract(slots, durationArg, options) {
        return createPlainDate(internal.movePlainDate(createMoveOps, true, slots, toDurationSlots(durationArg), options));
    },
    until(slots, otherArg, options) {
        return createDuration(internal.diffPlainDates(createDiffOps, false, slots, toPlainDateSlots(otherArg), options));
    },
    since(slots, otherArg, options) {
        return createDuration(internal.diffPlainDates(createDiffOps, true, slots, toPlainDateSlots(otherArg), options));
    },
    equals(slots, otherArg) {
        return internal.plainDatesEqual(slots, toPlainDateSlots(otherArg));
    },
    toString: internal.formatPlainDateIso,
    toJSON(slots) {
        return internal.formatPlainDateIso(slots);
    },
    toLocaleString(slots, locales, options) {
        const [format, epochMilli] = prepPlainDateFormat(locales, options, slots);
        return format.format(epochMilli);
    },
    toZonedDateTime(slots, options) {
        const optionsObj = !internal.isObjectLike(options) || options instanceof TimeZone
            ? { timeZone: options }
            : options;
        return createZonedDateTime(internal.plainDateToZonedDateTime(refineTimeZoneSlot, toPlainTimeSlots, createTimeZoneOps, slots, optionsObj));
    },
    toPlainDateTime(slots, plainTimeArg) {
        return createPlainDateTime(internal.plainDateToPlainDateTime(slots, optionalToPlainTimeFields(plainTimeArg)));
    },
    toPlainYearMonth(slots) {
        return createPlainYearMonth(internal.plainDateToPlainYearMonth(createYearMonthRefineOps, slots, this));
    },
    toPlainMonthDay(slots) {
        return createPlainMonthDay(internal.plainDateToPlainMonthDay(createMonthDayRefineOps, slots, this));
    },
    getISOFields: removeBranding,
    getCalendar: createCalendarFromSlots,
    valueOf: neverValueOf,
}, {
    from(arg, options) {
        return createPlainDate(toPlainDateSlots(arg, options));
    },
    compare(arg0, arg1) {
        return internal.compareIsoDateFields(toPlainDateSlots(arg0), toPlainDateSlots(arg1));
    },
});
// Utils
// -----------------------------------------------------------------------------
function toPlainDateSlots(arg, options) {
    options = internal.copyOptions(options);
    if (internal.isObjectLike(arg)) {
        const slots = (getSlots(arg) || {});
        switch (slots.branding) {
            case internal.PlainDateBranding:
                internal.refineOverflowOptions(options); // parse unused options
                return slots;
            case internal.PlainDateTimeBranding:
                internal.refineOverflowOptions(options); // parse unused options
                return internal.createPlainDateSlots(slots);
            case internal.ZonedDateTimeBranding:
                internal.refineOverflowOptions(options); // parse unused options
                return internal.zonedDateTimeToPlainDate(createTimeZoneOffsetOps, slots);
        }
        return internal.refinePlainDateBag(createDateRefineOps(getCalendarSlotFromBag(arg)), arg, options);
    }
    const res = internal.parsePlainDate(arg);
    internal.refineOverflowOptions(options); // parse unused options
    return res;
}

// Compound Adapter Functions
// -----------------------------------------------------------------------------
function fieldsAdapter(calendarProtocol, fieldsMethod, fieldNames) {
    return [...fieldsMethod.call(calendarProtocol, fieldNames)];
}
function mergeFieldsAdapter(calendarProtocol, mergeFields, fields, additionalFields) {
    return internal.requireObjectLike(mergeFields.call(calendarProtocol, Object.assign(Object.create(null), fields), Object.assign(Object.create(null), additionalFields)));
}
function dateFromFieldsAdapter(calendarProtocol, dateFromFields, fields, options) {
    return getPlainDateSlots(dateFromFields.call(calendarProtocol, Object.assign(Object.create(null), fields), options));
}
function yearMonthFromFieldsAdapter(calendarProtocol, yearMonthFromFields, fields, options) {
    return getPlainYearMonthSlots(yearMonthFromFields.call(calendarProtocol, Object.assign(Object.create(null), fields), options));
}
function monthDayFromFieldsAdapter(calendarProtocol, monthDayFromFields, fields, options) {
    return getPlainMonthDaySlots(monthDayFromFields.call(calendarProtocol, Object.assign(Object.create(null), fields), options));
}
function dateAddAdapter(calendarProtocol, dateAdd, isoFields, durationFields, options) {
    return getPlainDateSlots(dateAdd.call(calendarProtocol, createPlainDate(internal.createPlainDateSlots(isoFields, calendarProtocol)), createDuration(internal.createDurationSlots(durationFields)), options));
}
function dateUntilAdapter(calendarProtocol, dateUntil, isoFields0, isoFields1, largestUnit, origOptions) {
    return getDurationSlots(dateUntil.call(calendarProtocol, createPlainDate(internal.createPlainDateSlots(isoFields0, calendarProtocol)), createPlainDate(internal.createPlainDateSlots(isoFields1, calendarProtocol)), Object.assign(Object.create(null), origOptions, {
        largestUnit: internal.unitNamesAsc[largestUnit],
    })));
}
function dayAdapter(calendarProtocol, dayMethod, isoFields) {
    return internal.requirePositiveInteger(dayMethod.call(calendarProtocol, createPlainDate(internal.createPlainDateSlots(isoFields, calendarProtocol))));
}
// Compound Adapter Sets
// -----------------------------------------------------------------------------
const refineAdapters = { fields: fieldsAdapter };
const dateRefineAdapters = {
    dateFromFields: dateFromFieldsAdapter,
    ...refineAdapters,
};
const yearMonthRefineAdapters = {
    yearMonthFromFields: yearMonthFromFieldsAdapter,
    ...refineAdapters,
};
const monthDayRefineAdapters = {
    monthDayFromFields: monthDayFromFieldsAdapter,
    ...refineAdapters,
};
const modAdapters = { mergeFields: mergeFieldsAdapter };
const dateModAdapters = { ...dateRefineAdapters, ...modAdapters };
const yearMonthModAdapters = {
    ...yearMonthRefineAdapters,
    ...modAdapters,
};
const monthDayModAdapters = { ...monthDayRefineAdapters, ...modAdapters };
const moveAdapters = { dateAdd: dateAddAdapter };
const diffAdapters = { ...moveAdapters, dateUntil: dateUntilAdapter };
const yearMonthMoveAdapters = { ...moveAdapters, day: dayAdapter };
const yearMonthDiffAdapters = { ...diffAdapters, day: dayAdapter };
function createAdapterCompoundOps(calendarProtocol, adapterFuncs) {
    const keys = Object.keys(adapterFuncs).sort();
    const boundFuncs = {};
    for (const key of keys) {
        boundFuncs[key] = internal.bindArgs(adapterFuncs[key], calendarProtocol, calendarProtocol[key]);
    }
    return boundFuncs;
}

// Refine
const createYearMonthRefineOps = createCompoundOpsCreator(yearMonthRefineAdapters);
const createDateRefineOps = createCompoundOpsCreator(dateRefineAdapters);
const createMonthDayRefineOps = createCompoundOpsCreator(monthDayRefineAdapters);
// Mod
const createYearMonthModOps = createCompoundOpsCreator(yearMonthModAdapters);
const createDateModOps = createCompoundOpsCreator(dateModAdapters);
const createMonthDayModOps = createCompoundOpsCreator(monthDayModAdapters);
// Math
const createMoveOps = createCompoundOpsCreator(moveAdapters);
const createDiffOps = createCompoundOpsCreator(diffAdapters);
const createYearMonthMoveOps = createCompoundOpsCreator(yearMonthMoveAdapters);
const createYearMonthDiffOps = createCompoundOpsCreator(yearMonthDiffAdapters);
// -----------------------------------------------------------------------------
function createCompoundOpsCreator(adapterFuncs) {
    return (calendarSlot) => {
        if (typeof calendarSlot === 'string') {
            return internal.createNativeStandardOps(calendarSlot); // has everything
        }
        return createAdapterCompoundOps(calendarSlot, adapterFuncs);
    };
}

const [Duration, createDuration, getDurationSlots] = createSlotClass(internal.DurationBranding, internal.constructDurationSlots, {
    ...durationGetters,
    blank(slots) {
        return internal.queryDurationBlank(slots);
    },
    sign(slots) {
        return internal.queryDurationSign(slots);
    },
}, {
    with(slots, mod) {
        return createDuration(internal.durationWithFields(slots, mod));
    },
    add(slots, otherArg, options) {
        return createDuration(internal.addDurations(refinePublicRelativeTo, createDiffOps, createTimeZoneOps, false, slots, toDurationSlots(otherArg), options));
    },
    subtract(slots, otherArg, options) {
        return createDuration(internal.addDurations(refinePublicRelativeTo, createDiffOps, createTimeZoneOps, true, slots, toDurationSlots(otherArg), options));
    },
    negated(slots) {
        return createDuration(internal.negateDuration(slots));
    },
    abs(slots) {
        return createDuration(internal.absDuration(slots));
    },
    round(slots, options) {
        return createDuration(internal.roundDuration(refinePublicRelativeTo, createDiffOps, createTimeZoneOps, slots, options));
    },
    total(slots, options) {
        return internal.totalDuration(refinePublicRelativeTo, createDiffOps, createTimeZoneOps, slots, options);
    },
    toString: internal.formatDurationIso,
    toLocaleString(slots, locales, options) {
        return Intl.DurationFormat
            ? new Intl.DurationFormat(locales, options).format(this)
            : internal.formatDurationIso(slots);
    },
    toJSON(slots) {
        return internal.formatDurationIso(slots);
    },
    valueOf: neverValueOf,
}, {
    from(arg) {
        return createDuration(toDurationSlots(arg));
    },
    compare(durationArg0, durationArg1, options) {
        return internal.compareDurations(refinePublicRelativeTo, createDiffOps, createTimeZoneOps, toDurationSlots(durationArg0), toDurationSlots(durationArg1), options);
    },
});
// Utils
// -----------------------------------------------------------------------------
function toDurationSlots(arg) {
    if (internal.isObjectLike(arg)) {
        const slots = getSlots(arg);
        if (slots && slots.branding === internal.DurationBranding) {
            return slots;
        }
        return internal.refineDurationBag(arg);
    }
    return internal.parseDuration(arg);
}
function refinePublicRelativeTo(relativeTo) {
    if (relativeTo !== undefined) {
        if (internal.isObjectLike(relativeTo)) {
            const slots = (getSlots(relativeTo) || {});
            switch (slots.branding) {
                case internal.ZonedDateTimeBranding:
                case internal.PlainDateBranding:
                    return slots;
                case internal.PlainDateTimeBranding:
                    return internal.createPlainDateSlots(slots);
            }
            const calendar = getCalendarSlotFromBag(relativeTo); // !!!
            const res = internal.refineMaybeZonedDateTimeBag(refineTimeZoneSlot, createTimeZoneOps, createDateRefineOps(calendar), relativeTo);
            return { ...res, calendar };
        }
        return internal.parseZonedOrPlainDateTime(relativeTo);
    }
}

// Order matters. See call to createProtocolChecker
const calendarMethods = {
    toString(slots) {
        return slots.id;
    },
    toJSON(slots) {
        return slots.id;
    },
    ...calendarFieldMethods,
    dateAdd({ id, native }, plainDateArg, durationArg, options) {
        return createPlainDate(internal.createPlainDateSlots(native.dateAdd(toPlainDateSlots(plainDateArg), toDurationSlots(durationArg), options), id));
    },
    dateUntil({ native }, plainDateArg0, plainDateArg1, options) {
        return createDuration(internal.createDurationSlots(native.dateUntil(toPlainDateSlots(plainDateArg0), toPlainDateSlots(plainDateArg1), internal.refineCalendarDiffOptions(options))));
    },
    dateFromFields({ id, native }, fields, options) {
        return createPlainDate(internal.refinePlainDateBag(native, fields, options, internal.getRequiredDateFields(id)));
    },
    yearMonthFromFields({ id, native }, fields, options) {
        return createPlainYearMonth(internal.refinePlainYearMonthBag(native, fields, options, internal.getRequiredYearMonthFields(id)));
    },
    monthDayFromFields({ id, native }, fields, options) {
        return createPlainMonthDay(internal.refinePlainMonthDayBag(native, false, fields, options, internal.getRequiredMonthDayFields(id)));
    },
    fields({ native }, fieldNames) {
        /*
        Bespoke logic for converting Iterable to string[], while doing some validation
        */
        const allowed = new Set(internal.dateFieldNamesAlpha);
        const fieldNamesArray = [];
        for (const fieldName of fieldNames) {
            internal.requireString(fieldName);
            if (!allowed.has(fieldName)) {
                throw new RangeError(internal.forbiddenField(fieldName));
            }
            allowed.delete(fieldName); // prevents duplicates! can this be done somewhere else?
            fieldNamesArray.push(fieldName);
        }
        return native.fields(fieldNamesArray);
    },
    mergeFields({ native }, fields0, fields1) {
        return native.mergeFields(internal.excludeUndefinedProps(internal.requireNonNullish(fields0)), internal.excludeUndefinedProps(internal.requireNonNullish(fields1)));
    },
};
const [Calendar] = createSlotClass('Calendar', (id) => {
    const slotId = internal.resolveCalendarId(internal.requireString(id));
    const calendarNative = internal.createNativeStandardOps(slotId);
    return {
        branding: 'Calendar',
        id: slotId,
        native: calendarNative,
    };
}, {
    id(slots) {
        return slots.id;
    },
}, calendarMethods, {
    from(arg) {
        const calendarSlot = refineCalendarSlot(arg); // either string or CalendarProtocol
        return typeof calendarSlot === 'string'
            ? new Calendar(calendarSlot)
            : calendarSlot;
    },
});
function getCalendarSlotFromBag(bag) {
    return extractCalendarSlotFromBag(bag) || internal.isoCalendarId;
}
function extractCalendarSlotFromBag(bag) {
    const { calendar } = bag;
    if (calendar !== undefined) {
        return refineCalendarSlot(calendar);
    }
}
function refineCalendarSlot(arg) {
    if (internal.isObjectLike(arg)) {
        return ((getSlots(arg) || {}).calendar ||
            validateCalendarProtocol(arg));
    }
    return refineCalendarSlotString(arg);
}
function refineCalendarSlotString(arg) {
    return internal.resolveCalendarId(internal.parseCalendarId(internal.requireString(arg)));
}
const validateCalendarProtocol = createProtocolValidator(
// remove toString/toJSON/era/eraYear
Object.keys(calendarMethods).slice(4));

const Now = Object.defineProperties({}, {
    ...internal.createStringTagDescriptors('Temporal.Now'),
    ...internal.createPropDescriptors({
        timeZoneId() {
            return internal.getCurrentTimeZoneId(); // we call separately to return function.name
        },
        instant() {
            return createInstant(internal.createInstantSlots(internal.getCurrentEpochNanoseconds()));
        },
        zonedDateTime(calendar, timeZone = internal.getCurrentTimeZoneId()) {
            return createZonedDateTime(internal.createZonedDateTimeSlots(internal.getCurrentEpochNanoseconds(), refineTimeZoneSlot(timeZone), refineCalendarSlot(calendar)));
        },
        zonedDateTimeISO(timeZone = internal.getCurrentTimeZoneId()) {
            return createZonedDateTime(internal.createZonedDateTimeSlots(internal.getCurrentEpochNanoseconds(), refineTimeZoneSlot(timeZone), internal.isoCalendarId));
        },
        plainDateTime(calendar, timeZone = internal.getCurrentTimeZoneId()) {
            return createPlainDateTime(internal.createPlainDateTimeSlots(internal.getCurrentIsoDateTime(createTimeZoneOffsetOps(refineTimeZoneSlot(timeZone))), refineCalendarSlot(calendar)));
        },
        plainDateTimeISO(timeZone = internal.getCurrentTimeZoneId()) {
            return createPlainDateTime(internal.createPlainDateTimeSlots(internal.getCurrentIsoDateTime(createTimeZoneOffsetOps(refineTimeZoneSlot(timeZone))), internal.isoCalendarId));
        },
        plainDate(calendar, timeZone = internal.getCurrentTimeZoneId()) {
            return createPlainDate(internal.createPlainDateSlots(internal.getCurrentIsoDateTime(createTimeZoneOffsetOps(refineTimeZoneSlot(timeZone))), refineCalendarSlot(calendar)));
        },
        plainDateISO(timeZone = internal.getCurrentTimeZoneId()) {
            return createPlainDate(internal.createPlainDateSlots(internal.getCurrentIsoDateTime(createTimeZoneOffsetOps(refineTimeZoneSlot(timeZone))), internal.isoCalendarId));
        },
        plainTimeISO(timeZone = internal.getCurrentTimeZoneId()) {
            return createPlainTime(internal.createPlainTimeSlots(internal.getCurrentIsoDateTime(createTimeZoneOffsetOps(refineTimeZoneSlot(timeZone)))));
        },
    }),
});

const Temporal = Object.defineProperties({}, {
    ...internal.createStringTagDescriptors('Temporal'),
    ...internal.createPropDescriptors({
        PlainYearMonth,
        PlainMonthDay,
        PlainDate,
        PlainTime,
        PlainDateTime,
        ZonedDateTime,
        Instant,
        Calendar,
        TimeZone,
        Duration,
        Now,
    }),
}); // !!! (for tests)

exports.DateTimeFormat = DateTimeFormat;
exports.Temporal = Temporal;
exports.toTemporalInstant = toTemporalInstant;
